/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_user.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <stdio.h>
#include <string.h>
#include "h.h"

extern	char	translate_esc(char *, char *);

char	strbuff[100];

#define KLINE_RET_PERM 2
#define KLINE_RET_DELOK 1
#define KLINE_DEL_ERR 0


/*
** m_rehashall - WonKnet
** when sent by a Services Bot will rehash all servers
**
*/                
int     m_rehashall(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int     parc;
char    *parv[]; 
{
    if (!AUSTservices(cptr,sptr))
    {
        sendto_one(sptr, err_str(ERR_NOULINE), me.name, parv[0]);
        return 0;
    }
    else
	sendto_serv_butone(cptr, ":%s REHASHALL %s", parv[0], parv[1]);
	(void)rehash(&me, &me, 2);
} 

char
strlower(char *buf)
{
    int     i = 0;

    while (buf[i] != '\0')
    {
	buf[i] = tolower(buf[i]);
	i++;
    }
    buf[i] = '\0';
}

char   *
thedate()
{
    time_t  timer;

    time(&timer);
    strcpy(strbuff, ctime(&timer));
    strbuff[25] = 0;
    return strbuff;
}
 
/* If nokick, call above function */
int 
prepare_transport_message(sptr, acptr, notice, from, cmd, nick, msg)
    aClient *sptr, *acptr;
    int     notice;
    char   *from, *cmd, *nick, *msg;
{
    if (!notice && MyConnect(sptr) && acptr->user && acptr->user->away)
	sendto_one(sptr, rpl_str(RPL_AWAY), me.name, from, acptr->name,
		   acptr->user->away);
#if 0
    if (IsNoKick(acptr) && MyConnect(sptr))
    {
	if (!notice)
	{
	    parse_servicemsg(acptr, sptr, from, cmd, nick, msg);
	}
    }
    else {
#endif
/* Slight Kludge here for changing PRIVMSG into P and NOTICE into NO 
 * -- TQ */
        if(IsServer(acptr->from) && Protocol(acptr->from) >= 10) {
          if(*cmd == 'P') cmd = "P";
          if(*cmd == 'N') cmd = "NO";
        }
	sendto_prefix_one(acptr, sptr, ":%s %s %s :%s",
			  from, cmd, nick, msg);
#if 0
    }
#endif
    return 0;
}

/* Allow uline clients to do modes */
int 
IsUline(cptr, sptr)
    aClient *cptr, *sptr;
{
    aConfItem *aconf;

    if ((aconf = find_conf_host(cptr->confs, sptr->name, CONF_UWORLD)) ||
	(sptr->user && (aconf = find_conf_host(cptr->confs, sptr->user->server->name, CONF_UWORLD))))
	return 1;
    return 0;
}

/* m_gnotice  (Russell) sort of like wallop, but only to +g clients on
** this server.		-- Thanks to DALnet IRCD Distribution
**      parv[0] = sender prefix
**      parv[1] = message text
*/
int 
m_gnotice(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    char   *message, *pv[4];

    message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message))
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "GNOTICE");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr))
    {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0;
    }
    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GNOTICE :%s",
		       parv[0], message);
    sendto_failops("NOTICE from %s: %s", parv[0], message);
    return 0;
}

/*
** m_globops (write to opers who are +g currently online)	-- DALnet Distrub
**      parv[0] = sender prefix
**      parv[1] = message text
*/
int 
m_globops(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    char   *message, *pv[4];

    message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message))
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "GLOBOPS");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr) && !IsAnOper(cptr))
    {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0;
    }
    sendto_serv_butone(IsServer(cptr) ? cptr : NULL,
		       ":%s GLOBOPS :%s", parv[0], message);
    sendto_failops_whoare_opers("OPER from %s: %s", parv[0], message);
    return 0;
}

/* m_goper  (Russell) sort of like wallop, but only to ALL +o clients on
** every server.	--DALnet Distrib
**      parv[0] = sender prefix
**      parv[1] = message text
*/
int 
m_goper(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    char   *message, *pv[4];

    message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message))
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "GLOBAL");
	return 0;
    }
/*      if (!IsServer(sptr) && MyConnect(sptr) && !IsAnOper(sptr))*/
    if (!IsServer(sptr) && !AUSTservices(cptr, sptr))
    {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return 0;
    }
    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GOPER :%s",
		       parv[0], message);
    sendto_opers("from %s: %s", parv[0], message);
    return 0;
}

#ifdef PROFILE_BUILD
/*
 * m_prof - dump profiling info (Linux/glibc ONLY! - DL)
 *
 * This _must_ be built with -pg or will segfault on first call.
 */
int 
m_prof(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
	extern void _start (void), etext (void);

	if (!IsAnOper(cptr) || !MyClient(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	sendto_realops("Profile checkpoint by %s", parv[0]);
	_mcleanup();
	__monstartup(&_start, &etext);
	return 0;
}
#endif

/*
 * Add a temporary line to the configuration
 */
void 
add_temp_conf(status, host, passwd, name, port, class, temp)
    unsigned int status;
    char   *host;
    char   *passwd;
    char   *name;
    int     port, class, temp;	/* temp: 0 = perm 1 = temp */
{
    Reg1 aConfItem *aconf;

    aconf = make_conf();

    aconf->tmpconf = temp;
    aconf->status = status;
    if (host)
	DupString(aconf->host, host);
    if (passwd)
	DupString(aconf->passwd, passwd);
    if (name)
	DupString(aconf->name, name);
    aconf->port = port;
    if (class)
	Class(aconf) = find_class(class);
    if (!find_temp_conf_entry(aconf, status))
    {
	aconf->next = conf;
	conf = aconf;
    }
    aconf = NULL;

    if (aconf)
	free_conf(aconf);
}

/*
 * delete a temporary conf line.  *only* temporary conf lines may be deleted.
 */
int 
del_temp_conf(status, host, passwd, name, port, class)
    unsigned int status;
    char   *host;
    char   *passwd;
    char   *name;
    int     port, class;
{
    Reg1 aConfItem *aconf;
    Reg3 aConfItem *bconf;
    u_int   mask;
    u_int   result = KLINE_DEL_ERR;

    aconf = make_conf();

    aconf->status = status;
    if (host)
	DupString(aconf->host, host);
    if (passwd)
	DupString(aconf->passwd, passwd);
    if (name)
	DupString(aconf->name, name);
    aconf->port = port;
    if (class)
	Class(aconf) = find_class(class);
    mask = CONF_KILL;
    if (bconf = find_temp_conf_entry(aconf, mask))	/* only if non-null ptr */
    {
	if (bconf->tmpconf == KLINE_PERM)
	    return result = KLINE_RET_PERM;	/* Kline permanent */
	if (!IsIllegal(bconf))
	{
	    bconf->status |= CONF_ILLEGAL;	/* just mark illegal */
	    result = KLINE_RET_DELOK;	/* same as deletion */
	}
    }
    if (aconf)
	free_conf(aconf);
    return result;		/* if it gets to here, it doesn't exist */
}

/*
 * find_temp_conf_entry
 *
 * - looks for a match on all given fields for a TEMP conf line.
 *  Right now the passwd,port, and class fields are ignored, because it's
 *  only useful for k:lines anyway.  -Russell   11/22/95
 *  1/21/95 Now looks for any conf line.  I'm leaving this routine and its
 *  call in because this routine has potential in future upgrades. -Russell
 */
aConfItem *
find_temp_conf_entry(aconf, mask)
    aConfItem *aconf;
    u_int   mask;
{
    Reg1 aConfItem *bconf;

    for (bconf = conf, mask &= ~CONF_ILLEGAL; bconf; bconf = bconf->next)
    {
	if (!(bconf->status & mask) || (bconf->port != aconf->port))
	    continue;

/*                if (!bconf->tempconf) continue; */
	if ((BadPtr(bconf->host) && !BadPtr(aconf->host)) ||
	    (BadPtr(aconf->host) && !BadPtr(bconf->host)))
	    continue;
	if (!BadPtr(bconf->host) && mycmp(bconf->host, aconf->host))
	    continue;

/*                if ((BadPtr(bconf->passwd) && !BadPtr(aconf->passwd)) ||
                    (BadPtr(aconf->passwd) && !BadPtr(bconf->passwd)))
                        continue;
                if (!BadPtr(bconf->passwd) &&
                    mycmp(bconf->passwd, aconf->passwd))
                        continue; */

	if ((BadPtr(bconf->name) && !BadPtr(aconf->name)) ||
	    (BadPtr(aconf->name) && !BadPtr(bconf->name)))
	    continue;
	if (!BadPtr(bconf->name) && mycmp(bconf->name, aconf->name))
	    continue;
	break;
    }
    return bconf;
}


/*
** m_kline;
**      parv[0] = sender prefix
**      parv[1] = nickname
**      parv[2] = comment or filename
*/
int 
m_kline(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
#ifdef OPER_KLINE
    char   *host, *tmp, *hosttemp;
    char    uhost[80], name[80], reason[1024];
    int     ip1, ip2, ip3, temp;
    aClient *acptr;
    Reg2 int killflag;
    int     i;

    if (!MyClient(sptr) || !IsAnOper(sptr))
    {
#endif
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
#ifdef OPER_KLINE
    }

    if (check_registered(sptr))
	return 0;

    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "KLINE");
	return 0;
    }


/* This patch allows opers to quote kline by address as well as nick
 * --Russell
 */
    if (hosttemp = strchr(parv[1], '@'))
    {
	temp = 0;
	while (temp <= 20)
	    name[temp++] = 0;
	strcpy(uhost, ++hosttemp);
	strncpy(name, parv[1], hosttemp - 1 - parv[1]);
	if (name[0] == '\0' || uhost[0] == '\0')
	{
	    Debug((DEBUG_INFO, "KLINE: Bad field!"));
	    sendto_one(sptr, "NOTICE %s :If you're going to add a userhost, at LEAST specify both fields",
		       parv[0]);
	    return 0;
	}
	if (!strcmp(uhost, "*") || !strchr(uhost, '.'))
	{
	    sendto_one(sptr, "NOTICE %s :What a sweeping K:Line.  If only your admin knew you tried that..",
		       parv[0]);
	    return 0;
	}
    }

/* by nick */
    else
    {
	if (!(acptr = find_client(parv[1], NULL)))
	{
	    if (!(acptr = get_history(parv[1], (long) KILLCHASETIMELIMIT)))
	    {
		sendto_one(sptr, "NOTICE %s :Can't find user %s to add KLINE",
			   parv[0], parv[1]);
		return 0;
	    }
	}

	if (!acptr->user)
	    return 0;

	strcpy(name, acptr->user->username);
	if (MyClient(acptr))
	    host = acptr->sockhost;
	else
	    host = acptr->user->host;

	/* Sanity checks */

	if (name == '\0' || host == '\0')
	{
	    Debug((DEBUG_INFO, "KLINE: Bad field"));
	    sendto_one(sptr, "NOTICE %s :Bad field!", parv[0]);
	    return 0;
	}

	/* Add some wildcards */


	strcpy(uhost, host);
	if (strchr("0123456789", host[0]))
	{
	    if (sscanf(host, "%d.%d.%d.%*d", &ip1, &ip2, &ip3))
		sprintf(uhost, "%d.%d.%d.*", ip1, ip2, ip3);
	}
	else if (sscanf(host, "%*[^.].%*[^.].%s", uhost))
	{			/* Not really... */
	    tmp = (char *) strchr(host, '.');
	    sprintf(uhost, "*%s", tmp);
	}
    }

    sendto_ops("%s added a temp k:line for %s@%s %s", parv[0], name, uhost, parv[2] ? parv[2] : "");
    add_temp_conf(CONF_KILL, uhost, parv[2], name, 0, 0, 1);
    for (i = 0; i <= highest_fd; i++)
    {
	if (!(acptr = local[i]) || IsMe(acptr) || IsLog(acptr))
	    continue;
	killflag = IsPerson(acptr) ? find_kill(acptr) : 0;
        if (killflag)
        {
#ifdef P_LINES
                if (killflag == -3)
                        pline_quit(acptr);
		else
#endif
                if (IsAnOper(acptr) && killflag == -2) { 
                        sendto_ops("G-line matched but not activated for Oper %s",
                              get_client_name(acptr, FALSE));
                }
                else
                {
                   sendto_ops(killflag == -2 ? "G-line active for %s" :
                         "K-line active for %s", get_client_name(acptr,FALSE));
                   (void) exit_client(cptr, acptr, &me, killflag == -2 ? "G-lined" :
                              "K-lined");
                }
        }
    }

#endif
}


/*
 *  m_unkline
 *    parv[0] = sender prefix
 *    parv[1] = userhost
 */

int 
m_unkline(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{

    int     result, temp;
    char   *hosttemp = parv[1], host[80], name[80];

#ifdef OPER_KLINE
    if (!IsAnOper(cptr) || !MyClient(sptr))
#endif
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
#ifdef OPER_KLINE
    if (parc < 2)
    {
	sendto_one(sptr, "NOTICE %s :Not enough parameters", parv[0]);
	return 0;
    }
    if (hosttemp = strchr(parv[1], '@'))
    {
	temp = 0;
	while (temp <= 20)
	    name[temp++] = 0;
	strcpy(host, ++hosttemp);
	strncpy(name, parv[1], hosttemp - 1 - parv[1]);
	if (name[0] == '\0' || host[0] == '\0')
	{
	    Debug((DEBUG_INFO, "UNKLINE: Bad field"));
	    sendto_one(sptr, "NOTICE %s : Both user and host fields must be non-null", parv[0]);
	    return 0;
	}
	result = del_temp_conf(CONF_KILL, host, NULL, name, NULL, NULL);
	if (result == KLINE_RET_PERM)
	{			/* Not a temporary line - result =2 */
	    sendto_one(sptr, "NOTICE %s :You may not remove permanent K:Lines - talk to the admin", parv[0]);
	    return 0;
	}
	if (result == KLINE_RET_DELOK)
	{			/* Successful result = 1 */
	    sendto_one(sptr, "NOTICE %s :Temp K:Line %s@%s is now removed.", parv[0], name, host);
	    sendto_ops("%s removed temp k:line %s@%s", parv[0], name, host);
	    return 0;
	}
	if (result == KLINE_DEL_ERR)
	{			/* Unsuccessful result = 0 */
	    sendto_one(sptr, "NOTICE %s :Temporary K:Line %s@%s not found", parv[0], name, host);
	    return 0;
	}
    }
#endif
}

/* List and skip all channels that are listen */
void 
list_next_channels(cptr, nr)
    aClient *cptr;
    int     nr;
{
    char    tempmode[12];
    char    formatted_buff[TOPICLEN + 15];
    aChannel *chptr = cptr->listing;

    chptr->mode.mode &= ~MODE_LISTED;
    while (is_listed(chptr) || nr > 0)
    {
	if (!IsOper(cptr))
	{
	    if (cptr->user && !(SecretChannel(chptr) && !IsMember(cptr, chptr)))
	    {
		if (IsListem(cptr))
		{
		    nr--;
		    sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name,
			  ShowChannel(cptr, chptr) ? chptr->chname : "*",
			       chptr->users, ShowChannel(cptr, chptr) ? chptr->topic : "");
		}
		else
		{
		    if (!(chptr->users == chptr->HasService))
		    {		/* Dont list chans which are 1 and has
				   services */
			nr--;
			sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name,
			  ShowChannel(cptr, chptr) ? chptr->chname : "*",
				   chptr->users, ShowChannel(cptr, chptr) ? chptr->topic : "");
		    }
		}
	    }
	}
/* Allow Opers to see all channels */
	else
	{
	    if (IsListem(cptr))
	    {
		nr--;
		strcpy(tempmode, "+");
		if (chptr->HasService > 0)
		    strcat(tempmode, "%");
		if (chptr->mode.mode & MODE_SECRET)
		    strcat(tempmode, "s");
		else if (chptr->mode.mode & MODE_PRIVATE)
		    strcat(tempmode, "p");
		if (chptr->mode.mode & MODE_MODERATED)
		    strcat(tempmode, "m");
		if (chptr->mode.mode & MODE_TOPICLIMIT)
		    strcat(tempmode, "t");
		if (chptr->mode.mode & MODE_INVITEONLY)
		    strcat(tempmode, "i");
		if (chptr->mode.mode & MODE_NOPRIVMSGS)
		    strcat(tempmode, "n");
		if (chptr->mode.limit)
		    strcat(tempmode, "l");
		if (*chptr->mode.key)
		    strcat(tempmode, "k");
		sprintf(formatted_buff, "%-10s %s", tempmode, chptr->topic);
		sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name,
			   chptr->chname,
			   chptr->users, formatted_buff);
	    }
	    else
	    {
		if (!(chptr->users == chptr->HasService))
		{		/* Dont list chans which are 1 and has
				   services */
		    nr--;
		    strcpy(tempmode, "+");
		    if (chptr->HasService > 0)
			strcat(tempmode, "%");
		    if (chptr->mode.mode & MODE_SECRET)
			strcat(tempmode, "s");
		    else if (chptr->mode.mode & MODE_PRIVATE)
			strcat(tempmode, "p");
		    if (chptr->mode.mode & MODE_MODERATED)
			strcat(tempmode, "m");
		    if (chptr->mode.mode & MODE_TOPICLIMIT)
			strcat(tempmode, "t");
		    if (chptr->mode.mode & MODE_INVITEONLY)
			strcat(tempmode, "i");
		    if (chptr->mode.mode & MODE_NOPRIVMSGS)
			strcat(tempmode, "n");
		    if (chptr->mode.limit)
			strcat(tempmode, "l");
		    if (*chptr->mode.key)
			strcat(tempmode, "k");
		    sprintf(formatted_buff, "%-10s %s", tempmode, chptr->topic);
		    sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name,
			       chptr->chname,
			       chptr->users, formatted_buff);
		}
	    }
	}
/* End of OPER SEE ALL *grin* */
	if (!(chptr = chptr->nextch))
	{
	    sendto_one(cptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
	    break;
	}
    }
    if ((cptr->listing = chptr))
	chptr->mode.mode |= MODE_LISTED;
}

/*
 * m_end_of_burst  - Added Xorath 6-14-96, rewritten by Run 24-7-96
 *               - and fixed by record and Kev 8/1/96
 * This the last message in a net.burst.
 * It clears a flag for the server sending the burst.
 *
 * parv[0] - sender prefix
 */
m_end_of_burst(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    if (!IsServer(sptr))
	return 0;

    sendto_ops("Completed receiving Burst Data from %s.", sptr->name);
    sendto_highprot_butone(sptr, 9, ":%s END_OF_BURST", parv[0]);

    return 0;
}

/* m_helpme		-- RogerY. Helper wallop
** parv[0] = sender prefix
** parv[1] = message text
*/      
int
m_helpme(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{   
    char   *message, *pv[4];
    
    message = parc > 1 ? parv[1] : NULL; 
    
    if (BadPtr(message))  
    {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
                   me.name, parv[0], "HELPME");
        return 0;
    }
    if (!IsServer(sptr) && !AUSTservices(cptr, sptr) && !IsHelper(sptr) && !IsAnOper(sptr))
    {
	sendto_one(cptr, err_str(ERR_NOTALLOWED), me.name, parv[0]);
	return 0;
    }
    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s HELPME :%s",
                       parv[0], message);
    sendto_helpers("from %s: %s", parv[0], message);
    return 0;
}   


/* m_helper		-- RogerY Give someone umode +h
** parv[0] = sender prefix
** parv[1] = username to change mode for
** parv[2] = Helper type
*/
int 
m_helper(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    Reg1 aClient *from;

    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "HELPER");
	return 0;
    }

    if (AUSTservices(cptr, sptr) || IsServer(sptr))
    {
	if (!(from = find_client(parv[1], (aClient *) NULL)))
	{
	    sendto_one(sptr, err_str(ERR_NOSUCHNICK),
		       me.name, parv[0], parv[1]);
	    return 0;
	}

	from->umodes |= FLAGS_HELPER;
	from->user->helpstatus = (parc < 3) ? 0 : atoi(parv[2]);
	if ((from->user->helpstatus > 256) ||
	    (from->user->helpstatus < 0))
	    from->user->helpstatus = 0;
	sendto_serv_butone(sptr, ":%s HELPER %s %s", 
	    me.name, parv[1], (parc<3)?"0":parv[2]);
	if (MyConnect(from)) {
	    sendto_one(from, rpl_str(RPL_YOURHELPER), me.name, parv[1], parv[2]);
            sendto_one(from, ":%s MODE %s :+h", parv[1], parv[1]);
        }
    }
    else
    {
	sendto_one(sptr, err_str(ERR_NOULINE), me.name, parv[0]);
	return 0;
    }
}

/* m_vremover		-- DL Virus Removers (based on m_helper)
** parv[0] = sender prefix
** parv[1] = username to change mode for
** parv[2] = Helper type
*/
int 
m_vremover(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    Reg1 aClient *from;

    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "VREMOVER");
	return 0;
    }

    if (AUSTservices(cptr, sptr) || IsServer(sptr))
    {
	if (!(from = find_client(parv[1], (aClient *) NULL)))
	{
	    sendto_one(sptr, err_str(ERR_NOSUCHNICK),
		       me.name, parv[0], parv[1]);
	    return 0;
	}

	from->umodes |= FLAGS_VREMOVER;
	sendto_highprot_butone(sptr, 13, ":%s VREMOVER %s", me.name, parv[1]);
	if (MyConnect(from)) {
	    sendto_one(from, rpl_str(RPL_YOURVREMOVER), me.name, parv[1]);
            sendto_one(from, ":%s MODE %s :+r", parv[1], parv[1]);
        }
    }
    else
    {
	sendto_one(sptr, err_str(ERR_NOULINE), me.name, parv[0]);
	return 0;
    }
}

/* m_virus		-- DL Distribute Virus Notice
** parv[0] = sender prefix
** parv[1] = entire message
*/
int 
m_virus(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int     parc;
    char   *parv[];
{
    if (parc < 1)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "V");
	return 0;
    }

    if (AUSTservices(cptr, sptr) || IsServer(sptr) || IsVRemover(sptr))
    {
	sendto_vremover_orig(": *** Virus -- from %s: %s", parv[0], parv[1]);
	return 0;
    }
    else
    {
	sendto_one(sptr, err_str(ERR_NOTALLOWED), me.name, parv[0]);
	return 0;
    }
}

/* 
**	-- TazQ		U: lines change someones nick
** parv[0] prefix
** parv[1] Nick to change
** parv[2] Nick to change to
*/

extern char sendanyways; /* in s_user.c, for allowing modnick to bypass NICK_DELAY */

int     m_modnick(cptr, sptr, parc, parv)
aClient *sptr, *cptr;
int     parc;
char    *parv[];
{
  aClient *acptr, *acptr2;
  char *temp[3];

  if (IsServer(cptr)) {
    if (find_conf_host(cptr->confs, sptr->name, CONF_UWORLD)) {
      if (parc < 3) {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
                   "MODNICK");
        return 1;
      }

      sendto_alert("MODNICK from %s for %s to %s", parv[0], parv[1], parv[2]);
 
      acptr = find_client(parv[1], NULL);
      acptr2 = find_client(parv[2], NULL); 

      if(acptr == NULL) {
        sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
        return 1; 
      }

      if(acptr2 != NULL) { /* Nick is already in use, ignore it */
        sendto_one(sptr, err_str(ERR_NICKNAMEINUSE), me.name, parv[0], parv[2]);
        return 1;
      }

      if(IsServer(acptr) || IsMe(acptr)) { /* Can't change servers nick */
        return 0;
      }

      if(MyConnect(acptr)) {
        temp[0] = parv[1];
        temp[1] = parv[2];
        temp[2] = NULL;
        sendanyways = 1;
        m_nick(acptr, acptr, 2, temp);
	sendanyways = 0;
      }
      else {
        sendto_one(acptr, ":%s %s %s %s", parv[0], 
		Protocol(acptr->user->server)>=12?"D":"MODNICK", 
		parv[1], parv[2]);
      }
    }
  }
  return 0;
}


/*
** RogerY -- ircd.links has the C/N lines in there. Services can add remove.
**   and saved to this file to effectivly make permanent.
**
** Scot@wonk.net -added functionality to write the Q: lines.
*/
int
write_ircdlinks()
{
    Reg1	aConfItem	*ac;
    FILE	*fp;

    rename(CNPATH, CNPATH".bak");

    if ((fp = fopen(CNPATH, "w")) == NULL)
    {
	sendto_ops("Error opening '%s' for writing", CNPATH);
	return 0;
    }

    fprintf(fp, "# %s automatically written on %s\n\n", 
	CNPATH, thedate());

    for (ac = conf; ac; ac = ac->next)
    {
	if (ac->status == CONF_CONNECT_SERVER)
	{
	    fprintf(fp, "C:%s:%s:%s:%d:%d\n",
		((ac->host[0]=='*')&&(ac->host[1]=='@'))?&ac->host[2]:ac->host,
		ac->passwd, ac->name, ac->port, get_conf_class(ac));
	}
	else if (ac->status == CONF_NOCONNECT_SERVER)
        {
            fprintf(fp, "N:%s:%s:%s:%d:%d\n",
		((ac->host[0]=='*')&&(ac->host[1]=='@'))?&ac->host[2]:ac->host,
                ac->passwd, ac->name, ac->port, get_conf_class(ac));    
	}
	else if (ac->status == CONF_VLINES)
	    fprintf(fp, "V:%s\n", ac->host);
	else if (ac->status == CONF_QUARANTINED_SERVER)
	{
	    fprintf(fp, "Q:*:%s:%s\n", ac->passwd, ac->name);
	}	

    }

    fclose(fp);
    return 0;
}

/*
** DL - ircd.alines - stores current A-lines
** (taboo DCC file-name endings/Anti-Virus lines)
*/
int
write_alines()
{
    aAline	*aaline;
    FILE	*fp;

    rename(APATH, APATH".bak");

    if ((fp = fopen(APATH, "w")) == NULL)
    {
	sendto_ops("Error opening '%s' for writing", APATH);
	return 0;
    }

    fprintf(fp, "# %s automatically written on %s\n\n",
	APATH, thedate());

    for (aaline = aline; aaline; aaline = aaline->next)
    {
        /* > 126 chars? tough, truncation occurs */
	fprintf(fp, "%.126s\n", aaline->pattern);
    }

    fclose(fp);
    return 0;
}

int
read_alines()
{
    aAline	*aaline;
    char	buf[128];
    int		len;
    FILE	*fp;

    /* Free current alines */
    if (aline) {
	while(aline->next) {
	  aaline = aline->next;
	  free(aline->pattern);
	  free(aline);
	  aline = aaline;
	}
        free(aline->pattern);
	free(aline);
	aline = NULL;
    }

    if ((fp = fopen(APATH, "r")) == NULL)
    {
	sendto_ops("Error opening '%s' for reading", APATH);
	return 0;
    }

    /* > 126 chars? tough, multiple rules are formed. */
    while(fgets(buf, 128, fp)) {
	if(buf[0] == '#')
		continue;
	if((len = strlen(buf)) && buf[len-1] == '\n')
		buf[len-1] = '\0';
	if(buf[0] == '\0')
		continue;
	collapse(buf);
        if(!aline) {
          aline = aaline = (struct Aline *)MyMalloc(sizeof(aAline));
        } else {
          aaline->next = (struct Aline *)MyMalloc(sizeof(aAline));
	  aaline = aaline->next;
        }
        DupString(aaline->pattern, buf);
        aaline->next = NULL;
    }

    fclose(fp);
    return 0;
}
